import tempfile
from pathlib import Path

import pytest

from rumydata.field import Text, Integer, Date, Choice


@pytest.fixture()
def tmpdir():
    with tempfile.TemporaryDirectory() as d:
        yield Path(d)


@pytest.fixture()
def basic() -> dict:
    return {
        'col1': Text(1),
        'col2': Integer(1),
        'col3': Date(),
        'col4': Choice(['X', 'Y', 'Z'])
    }



"""
This is a giant dumpster fire. It needs to be broken apart into a set of more
sensible scripts.
"""

import csv
import uuid
from pathlib import Path
from textwrap import dedent

import pytest
from openpyxl import Workbook

import rumydata.rules.cell
import rumydata.rules.cell as clr
import rumydata.table
from rumydata import exception as ex
from rumydata import field
from rumydata.rules import column as cr, table as tr, header as hr
from rumydata.table import CsvFile, ExcelFile, Layout


def write_row(directory, columns: rumydata.table.Layout, row, rows=False):
    p = Path(directory, str(uuid.uuid4()))
    with p.open('w', newline='') as o:
        writer = csv.writer(o)
        writer.writerow(list(columns.layout))
        if rows:
            for r in row:
                writer.writerow(r)
        else:
            writer.writerow(row)
    return p


@pytest.fixture()
def basic_good(tmpdir):
    p = Path(tmpdir, uuid.uuid4().hex)
    with p.open('w', newline='') as o:
        writer = csv.writer(o)
        writer.writerow(['col1', 'col2', 'col3', 'col4'])
        writer.writerow(['A', '1', '2020-01-01', 'X'])
    yield p.as_posix()


@pytest.fixture()
def basic_good_with_empty(tmpdir):
    p = Path(tmpdir, uuid.uuid4().hex)
    with p.open('w', newline='') as o:
        writer = csv.writer(o)
        writer.writerow(['col1', 'col2', 'col3', '', 'col4'])
        writer.writerow(['A', '1', '2020-01-01', '', 'X'])
    yield p.as_posix()


@pytest.fixture()
def basic_good_excel(tmpdir):
    p = Path(tmpdir, 'good.xlsx')
    wb = Workbook()
    ws = wb.active
    ws.append(['col1', 'col2', 'col3', 'col4'])
    ws.append(['A', '1', '2020-01-01', 'X'])
    wb.save(p)
    yield p.as_posix()


@pytest.fixture()
def basic_row_skip_good(tmpdir):
    p = Path(tmpdir, 'good.csv')
    with p.open('w', newline='') as o:
        writer = csv.writer(o)
        writer.writerow(['garbage'])
        writer.writerow(['garbage'])
        writer.writerow(['col1', 'col2', 'col3', 'col4'])
        writer.writerow(['A', '1', '2020-01-01', 'X'])
    yield p.as_posix()


@pytest.fixture()
def readme_layout():
    return {
        'col1': field.Text(8),
        'col2': field.Choice(['x', 'y', 'z'], nullable=True),
        'col3': field.Integer(1)
    }


@pytest.fixture()
def readme_data(tmpdir):
    p = Path(tmpdir, 'bobs_data.csv')
    p.write_text('\n'.join([
        "col1,col2,col3",
        "abc,x,-1",
        "def,,0",
        "ghi,a,1"
    ]))
    yield p.as_posix()


def empty_rows(rows, directory):
    p = Path(directory, str(uuid.uuid4()))
    with p.open('w', newline='') as f:
        w = csv.writer(f)
        w.writerow(['x'])
        for i in range(rows):
            w.writerow('')
    return p


def test_file_not_exists(basic):
    assert CsvFile(rumydata.table.Layout(basic)). \
        _has_error('abc123.csv', ex.FileError)


def test_file_good(basic_good, basic):
    assert not CsvFile(rumydata.table.Layout(basic)).check(basic_good)


def test_file_excel_good(basic_good_excel, basic):
    assert not ExcelFile(rumydata.table.Layout(basic)).check(basic_good_excel)


def test_file_invalid_type(basic):
    with pytest.raises(TypeError):
        CsvFile(rumydata.table.Layout(basic), file_type='xxx')


def test_file_row_skip_good(basic_row_skip_good, basic):
    assert not CsvFile(rumydata.table.Layout(basic), skip_rows=2).check(basic_row_skip_good)


def test_layout_good(basic, basic_good):
    assert not CsvFile(rumydata.table.Layout(basic)).check(basic_good)


def test_readme_example(readme_layout, readme_data):
    rex = rumydata.rules.cell.Choice.rule_exception()
    assert CsvFile(rumydata.table.Layout(readme_layout))._has_error(readme_data, rex)


@pytest.mark.parametrize('rows,me', [
    (1, 0),
    (2, 1),
    (101, 100),
    (int(1e5), 100),
])
def test_has_max_error(tmpdir, rows, me):
    fields = rumydata.table.Layout({'x': field.Field()})
    file = empty_rows(rows, tmpdir)
    assert CsvFile(fields, max_errors=me)._has_error(file, tr.MaxError.rule_exception())


@pytest.mark.parametrize('rows,me', [
    (1, 1),
    (99, 100),
    (100, 100),
    (100, int(1e5)),
])
def test_missing_max_error(tmpdir, rows, me):
    fields = rumydata.table.Layout({'x': field.Field()})
    file = empty_rows(rows, tmpdir)
    assert not CsvFile(fields, max_errors=me)._has_error(file, tr.MaxError.rule_exception())


def test_column_compare_row_good():
    fields = rumydata.table.Layout({
        'a': rumydata.field.Integer(1, rules=[rumydata.rules.cell.GreaterThanColumn('b')]),
        'b': rumydata.field.Integer(1)
    })
    assert not fields.check_row(['3', '2'])


@pytest.mark.parametrize('compare_rule,row', [
    (rumydata.rules.cell.GreaterThanColumn('x'), ['2', '3']),
])
def test_column_compare_file_good(tmpdir, compare_rule, row):
    cols = rumydata.table.Layout({'x': field.Field(), 'y': field.Field(rules=[compare_rule])})
    assert not CsvFile(cols).check(write_row(tmpdir, cols, row))


@pytest.mark.parametrize('compare_rule,row', [
    (rumydata.rules.cell.GreaterThanColumn('x'), ['1', '1']),
])
def test_column_compare_file_bad(tmpdir, compare_rule, row):
    cols = rumydata.table.Layout({'x': field.Field(), 'y': field.Field(rules=[compare_rule])})
    assert CsvFile(cols)._has_error(write_row(tmpdir, cols, row), compare_rule.rule_exception())


def test_header_file_bad(tmpdir):
    cols1 = rumydata.table.Layout({'x': field.Field()})
    cols2 = rumydata.table.Layout({'y': field.Field()})
    row = ['1']
    fp = write_row(tmpdir, cols2, row)
    assert CsvFile(cols1)._has_error(fp, hr.NoMissing.rule_exception())


def test_unique_bad(tmpdir):
    cols = rumydata.table.Layout({'x': field.Field(rules=[cr.Unique()])})
    f = write_row(tmpdir, cols, [['1'], ['1'], ['1']], rows=True)
    assert CsvFile(cols)._has_error(f, cr.Unique.rule_exception())


def test_unique_good(tmpdir):
    cols = rumydata.table.Layout({'x': field.Field(rules=[cr.Unique()])})
    f = write_row(tmpdir, cols, [['1'], ['2'], ['3']], rows=True)
    assert not CsvFile(cols).check(f)


@pytest.mark.parametrize('row,kwargs', [
    (['1', '1'], {}),
    (['1', '1'], dict(empty_row_ok=False)),
    (['', ''], dict(empty_row_ok=True))
])
def test_empty_row_good(row, kwargs):
    lay = rumydata.table.Layout({'x': field.Integer(1), 'y': field.Integer(2)}, **kwargs)
    assert not lay.check_row(row)


def test_empty_row_file_good(tmpdir):
    cols = rumydata.table.Layout({'x': field.Field()}, empty_row_ok=True)
    f = write_row(tmpdir, cols, [['1'], ['2'], ['']], rows=True)
    assert not CsvFile(cols).check(f)


@pytest.mark.parametrize('row', [
    (['1', '']),
    (['', '']),
    (['1', '1'])
])
def test_ignore_row(row):
    """ Test that ignore rows count as empty for the purpose of accepting empty rows """
    lay = rumydata.table.Layout(
        {'x': field.Ignore(), 'y': field.Integer(1)}, empty_row_ok=True
    )
    assert not lay.check_row(row)


def test_debug_mode_pre_process(mocker):
    """ Debug messages should only appear when debug method has been patched """
    try:
        # noinspection PyTypeChecker
        field.Date().check_cell(None)
    except AssertionError as ae:
        assert '[DEBUG]' not in str(ae)

    mocker.patch('rumydata.exception.debug', return_value=True)
    try:
        # noinspection PyTypeChecker
        field.Date().check_cell(None)
    except AssertionError as ae:
        assert '[DEBUG]' in str(ae)


def test_debug_mode(mocker):
    """ Debug messages should only appear when debug method has been patched """

    r = clr.make_static_cell_rule(lambda x: 1 / 0, 'raise an exception')
    try:
        # noinspection PyTypeChecker
        field.Field(rules=[r]).check_cell('1')
    except AssertionError as ae:
        assert '[DEBUG]' not in str(ae)

    mocker.patch('rumydata.exception.debug', return_value=True)
    try:
        # noinspection PyTypeChecker
        field.Field(rules=[r]).check_cell('1')
    except AssertionError as ae:
        assert '[DEBUG]' in str(ae)


def test_cell_trim():
    assert not field.Choice(['x'], strip=True).check_cell(' x ')


def test_column_trim():
    assert not field.Choice(['x'], strip=True).check_column([' x '])


def test_empty_column_good(basic, basic_good, basic_good_with_empty):
    assert not CsvFile(Layout(basic, empty_cols_ok=True)).check(basic_good_with_empty)
    assert not CsvFile(Layout(basic, empty_cols_ok=True)).check(basic_good)


def test_empty_column_bad(basic, basic_good_with_empty):
    with pytest.raises(AssertionError):
        assert CsvFile(Layout(basic, empty_cols_ok=False)).check(basic_good_with_empty)


@pytest.fixture()
def good_complex_file(tmpdir):
    p = Path(tmpdir, uuid.uuid4().hex)
    p.write_text(dedent("""
    c1,c2,c3,c4xyz,,c5,c6
    A,1,2020-01-01,X,,a,
    ,,,,,,
    B,2,2020-01-02,y,,a,
    """))
    yield p.as_posix()


def test_complex_good(good_complex_file):
    lay = Layout({
        'c1': field.Text(1, rules=[clr.NotNullIfCompare('c2')]),
        'c2': field.Integer(1, rules=[clr.OtherCantExist('c6')]),
        'c3': field.Date(rules=[clr.OtherMustExist('c4')]),
        'c4': field.Choice(['x', 'y', 'z'], case_insensitive=True),
        'c5': field.Text(1),
        'c6': field.Text(1, nullable=True)
    }, empty_row_ok=True, empty_cols_ok=True, header_mode='startswith')
    assert not CsvFile(lay, skip_rows=1).check(good_complex_file)


@pytest.fixture()
def bad_complex_file(tmpdir):
    p = Path(tmpdir, uuid.uuid4().hex)
    p.write_text(dedent("""
    c1,c2,c3,c4xyz,,c5,c6
    ,1,2020-01-01,,,a,a
    ,,,,,,
    B,2,2020-01-02,y,,a,
    """))
    yield p.as_posix()


def test_complex_bad(bad_complex_file):
    lay = Layout({
        'c1': field.Text(1, rules=[clr.NotNullIfCompare('c2')]),
        'c2': field.Integer(1, rules=[clr.OtherCantExist('c6')]),
        'c3': field.Date(rules=[clr.OtherMustExist('c4')]),
        'c4': field.Choice(['x', 'y', 'z'], case_insensitive=True),
        'c5': field.Text(1),
        'c6': field.Text(1, nullable=True)
    }, empty_row_ok=True, empty_cols_ok=True, header_mode='startswith')
    expected_ex = ['NotNullIfCompare',
                   'OtherCantExist', 'OtherMustExist']
    try:
        CsvFile(lay, skip_rows=1).check(bad_complex_file)
    except AssertionError as ex:
        print(str(ex))
        assert all(x in str(ex) for x in expected_ex)



"""
rumydata row rules

These rules are applied to header rows and are generally not intended to be
used directly. These rules ensure that the headers are named as expected, is
not missing any names, or containing any extras.
"""
from collections import namedtuple
from typing import List

from rumydata._base import _BaseRule

# this named tuple is here to allow for setting the default argument without
# needing to import the Layout class, which results in a circular import
_default_thing = namedtuple('DefaultDict', ['layout', 'header_mode', 'empty_cols_ok'])


class Rule(_BaseRule):
    """ Header Rule """

    _default_args = (_default_thing({}, 'exact', False),)

    def __init__(self, columns):
        super().__init__()
        self.header_mode = columns.header_mode
        self.definition = columns.layout
        self.empty_cols_ok = columns.empty_cols_ok

    def _prepare(self, data: List[str]) -> tuple:
        if self.empty_cols_ok:
            return [x for x in data if x != ''],
        else:
            return data,


class NoExtra(Rule):
    """ No extra header elements Rule """

    def _evaluator(self):
        modes = {
            'exact': lambda x: all([y in self.definition for y in x]),
            'startswith': lambda x: all([any([y.startswith(z) for z in self.definition]) for y in x]),
            'contains': lambda x: all([any([z in y for z in self.definition]) for y in x])
        }
        return modes[self.header_mode]

    def _explain(self):
        return 'Header row must not have unexpected columns'


class NoMissing(Rule):
    """ No missing header elements Rule """

    def _evaluator(self):
        modes = {
            'exact': lambda x: all([y in x for y in self.definition]),
            'startswith': lambda x: all([any([z.startswith(y) for z in x]) for y in self.definition]),
            'contains': lambda x: all([any([y in z for z in x]) for y in self.definition])
        }
        return modes[self.header_mode]

    def _explain(self) -> str:
        return 'Header row must not be missing any expected columns'


class NoDuplicate(Rule):
    """ No duplicate header elements Rule """

    def _evaluator(self):
        def starts_with(x):
            ixs = [[y.startswith(z) for z in list(self.definition.keys())].index(True) for y in x]
            return len(ixs) == len(set(ixs))

        def contains(x):
            ixs = [[z in y for z in list(self.definition.keys())].index(True) for y in x]
            return len(ixs) == len(set(ixs))

        modes = {
            'exact': lambda x: len(x) == len(set(x)),
            'startswith': starts_with,
            'contains': contains
        }
        return modes[self.header_mode]

    def _explain(self):
        return 'Header row must not contain duplicate values'


class ColumnOrder(Rule):
    """ Fixed header element order Rule """

    def _evaluator(self):
        def starts_with(x):
            ixs = [[y.startswith(z) for z in list(self.definition.keys())].index(True) for y in x]
            return ixs == sorted(ixs)

        def contains(x):
            ixs = [[z in y for z in list(self.definition.keys())].index(True) for y in x]
            return ixs == sorted(ixs)

        modes = {
            'exact': lambda x: x == list(self.definition),
            'startswith': starts_with,
            'contains': contains
        }
        return modes[self.header_mode]

    def _explain(self):
        return 'Header row must explicitly match order of definition'



import pytest

from rumydata.rules.column import *
from rumydata.rules.column import Rule


def recurse_subclasses(class_to_recurse):
    def generator(x):
        for y in x.__subclasses__():
            for z in generator(y):
                yield z
        yield x

    return list(generator(class_to_recurse))


@pytest.mark.parametrize('rule', recurse_subclasses(Rule))
def test_rule_prepare(rule):
    """
    All column preparation must accept a tuple of a value and a dictionary with
    comparison values that may be required.
    """
    r = rule(*rule._default_args)
    prep = r._prepare((['x', 'y']))
    assert isinstance(prep, tuple)
    assert isinstance(prep[0], list)


@pytest.mark.parametrize('value,expected', [
    (['1', '2', '3'], True),
    (['1', '1', '3'], False),
    (['', '', '1'], True),
    (['', '1', '1'], False)
])
def test_unique(value: list, expected: bool):
    r = Unique()
    assert r._evaluator()(*r._prepare(value)) is expected



import pytest

from rumydata.rules.cell import *
from rumydata.rules.cell import Rule


def recurse_subclasses(class_to_recurse):
    def generator(x):
        for y in x.__subclasses__():
            for z in generator(y):
                yield z
        yield x

    return list(generator(class_to_recurse))


@pytest.mark.parametrize('rule', recurse_subclasses(Rule))
def test_rule_prepare(rule):
    """
    All cell preparation must accept a tuple of a value and a dictionary with
    comparison values that may be required.
    """
    r = rule(*rule._default_args)
    assert isinstance(r._prepare(('1', {'x': '0'})), tuple)


@pytest.mark.parametrize('rule', recurse_subclasses(Rule))
def test_rule_evaluator_callable(rule):
    """ All rules must return a callable function """
    assert callable(rule(*rule._default_args)._evaluator())


@pytest.mark.parametrize('value,expected', [
    (' ', True),
    ('1', True),
    ('0', True),
    ('False', True),
    ('', False),
])
def test_not_null(value: str, expected: bool):
    r = NotNull()
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('length,value,expected', [
    (1, 'a', True),
    (2, 'a', False),
    (0, 'a', False)
])
def test_exact_char(value: str, expected: bool, length: int):
    r = ExactChar(length)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('length,value,expected', [
    (1, 'a', True),
    (2, 'aa', True),
    (0, 'a', True),
    (2, 'a', False),
    (3, 'aa', False)
])
def test_min_char(value: str, expected: bool, length: int):
    r = MinChar(length)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('length,value,expected', [
    (1, 'a', True),
    (2, 'aa', True),
    (0, 'a', False),
    (2, 'a', True),
    (3, 'aa', True)
])
def test_max_char(value: str, expected: bool, length: int):
    r = MaxChar(length)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('value,expected', [
    ('A', True),
    ('aa', True),
    ("\u0394", False)
])
def test_ascii_char(value: str, expected: bool):
    r = AsciiChar()
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('choice,data,expected,kwargs', [
    (['x'], 'x', True, {}),
    (['x'], 'y', False, {}),
    (['x', 'y'], 'y', True, {}),
    (['x'], ('x', {}), True, {}),
    (['X'], ('x', {}), True, dict(case_insensitive=True))
])
def test_choice(data, expected: bool, choice: list, kwargs):
    r = Choice(choice, **kwargs)
    assert r._evaluator()(*r._prepare(data)) is expected


@pytest.mark.parametrize('length,value,expected', [
    (1, 'a', False),
    (2, 'aa', False),
    (2, 'aa1', False),
    (0, 'a', True),  # if no digits are required, 'a' is valid
    (1, '1a', True),
    (2, '111a', True)
])
def test_min_digit(value: str, expected: bool, length: int):
    r = MinDigit(length)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('length,value,expected', [
    (1, 'a', True),
    (2, 'aa', True),
    (2, 'aa1', True),
    (0, 'a', True),  # if no digits are required, 'a' is valid
    (1, '1a', True),
    (2, '111a', False)
])
def test_max_digit(value: str, expected: bool, length: int):
    r = MaxDigit(length)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('value,expected', [
    ('123', True),
    ('123a', False),
    ('12.3', False)
])
def test_only_numbers(value: str, expected: bool):
    r = OnlyNumbers()
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('value,expected', [
    ('0123', False),
    ('1023', True),
    ('0', True),
    ('0.0', False)
])
def test_no_leading_zero(value: str, expected: bool):
    r = NoLeadingZero()
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('value,expected', [
    ('1', True),
    ('0', True),
    ('a', False),
])
def test_can_be_float(value: str, expected: bool):
    r = CanBeFloat()
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('value,expected', [
    ('1', True),
    ('0', True),
    ('0.0', False),
    ('0.1', False),
    ('a', False)
])
def test_can_be_integer(value: str, expected: bool):
    r = CanBeInteger()
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('decimals,value,expected', [
    (1, '1.0', True),
    (1, '123', True),
    (1, '0123', True),  # combine with NoLeadingZero to prevent this
    (1, '1.00', False),
    (2, '1.00', True)
])
def test_numeric_decimals(decimals: int, value: str, expected: bool):
    r = NumericDecimals(decimals)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('comparison,value,expected', [
    (1, 'x', False),
    (1, 'xx', True),
    (2, 'xx', False)
])
def test_length_gt(comparison: int, value: str, expected: bool):
    r = LengthGT(comparison)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('comparison,value,expected', [
    (1, 'x', True),
    (1, 'xx', True),
    (3, 'xx', False)
])
def test_length_gte(comparison: int, value: str, expected: bool):
    r = LengthGTE(comparison)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('comparison,value,expected', [
    (1, 'x', True),
    (2, 'xx', True),
    (3, 'xx', False)
])
def test_length_et(comparison: int, value: str, expected: bool):
    r = LengthET(comparison)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('comparison,value,expected', [
    (1, 'x', True),
    (2, 'xx', True),
    (3, 'xx', True),
    (0, 'x', False)
])
def test_length_lte(comparison: int, value: str, expected: bool):
    r = LengthLTE(comparison)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('comparison,value,expected', [
    (1, 'x', False),
    (2, 'xx', False),
    (3, 'xx', True),
    (0, 'x', False)
])
def test_length_lt(comparison: int, value: str, expected: bool):
    r = LengthLT(comparison)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('comparison,value,expected', [
    (1, '1', False),
    (1, '2', True),
    (1, '1.1', True),
    (0, '0', False)
])
def test_numeric_gt(comparison: float, value: str, expected: bool):
    r = NumericGT(comparison)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('comparison,value,expected', [
    (1, '1', True),
    (1, '2', True),
    (1, '1.1', True),
    (1, '1.0', True),
    (0, '0', True),
    (1, '0', False)
])
def test_numeric_gte(comparison: float, value: str, expected: bool):
    r = NumericGTE(comparison)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('comparison,value,expected', [
    (1, '1', True),
    (1, '2', False),
    (1, '1.1', False),
    (1, '1.0', True),
    (0, '0', True),
    (1, '0', False)
])
def test_numeric_et(comparison: float, value: str, expected: bool):
    r = NumericET(comparison)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('comparison,value,expected', [
    (1, '1', True),
    (1, '2', False),
    (1, '1.1', False),
    (1, '1.0', True),
    (0, '0', True),
    (1, '0', True)
])
def test_numeric_lte(comparison: float, value: str, expected: bool):
    r = NumericLTE(comparison)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('comparison,value,expected', [
    (1, '1', False),
    (1, '2', False),
    (1, '1.1', False),
    (1, '1.0', False),
    (1, '0.9', True),
    (0, '0', False),
    (1, '0', True)
])
def test_numeric_lt(comparison: float, value: str, expected: bool):
    r = NumericLT(comparison)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('value,expected,kwargs', [
    ('2020-01-01', True, {}),
    ('2020/01/01', False, {}),
    ('1901-01-01', True, {}),
    ('19010101', False, {}),
    ('9999-99-99', False, {}),
    ('2020-13-01', False, {}),
    ('2020-01-01', True, dict(truncate_time=True)),
    ('2020-01-01 00:00:00', True, dict(truncate_time=True)),
    (('2020-01-01 00:00:00', {}), True, dict(truncate_time=True)),
    ('2020-01-01 00:00:01', False, dict(truncate_time=True))
])
def test_can_be_date_iso(value: str, expected: bool, kwargs: dict):
    r = CanBeDateIso(**kwargs)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('comparison,value,expected', [
    ('2020-01-01', '2020-01-01', False),
    ('2020-01-01', '2020-01-02', True),
    ('2020-01-01', '2020-01-32', False),
    ('2020-01-01', '2019-12-31', False)
])
def test_date_gt(comparison: str, value: str, expected: bool):
    r = DateGT(comparison)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('comparison,value,expected', [
    ('2020-01-01', '2020-01-01', True),
    ('2020-01-01', '2020-01-02', True),
    ('2020-01-01', '2020-01-32', False),
    ('2020-01-01', '2019-12-31', False)
])
def test_date_gte(comparison: str, value: str, expected: bool):
    r = DateGTE(comparison)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('comparison,value,expected', [
    ('2020-01-01', '2020-01-01', True),
    ('2020-01-01', '2020-01-02', False),
    ('2020-01-01', '2020-01-32', False),
    ('2020-01-01', '2019-12-31', False)
])
def test_date_et(comparison: str, value: str, expected: bool):
    r = DateET(comparison)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('comparison,value,expected', [
    ('2020-01-01', '2020-01-01', True),
    ('2020-01-01', '2020-01-02', False),
    ('2020-01-01', '2020-01-32', False),
    ('2020-01-01', '2019-12-31', True)
])
def test_date_lte(comparison: str, value: str, expected: bool):
    r = DateLTE(comparison)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('comparison,value,expected', [
    ('2020-01-01', '2020-01-01', False),
    ('2020-01-01', '2020-01-02', False),
    ('2020-01-01', '2020-01-32', False),
    ('2020-01-01', '2019-12-31', True)
])
def test_date_lt(comparison: str, value: str, expected: bool):
    r = DateLT(comparison)
    assert r._evaluator()(*r._prepare(value)) is expected


@pytest.mark.parametrize('compared,data,expected', [
    ('x', ('1', {'x': '0'}), True),
    ('x', ('1', {'x': '1'}), False),
    ('z', ('1', {'x': '1', 'z': '0'}), True)
])
def test_greater_than_column(compared: str, data, expected: bool):
    r = GreaterThanColumn(compared)
    assert r._evaluator()(*r._prepare(data)) is expected


@pytest.mark.parametrize('compared,data,expected', [
    ('x', ('1', {'x': '0'}), True),
    ('x', ('1', {'x': '1'}), True),
    ('x', ('1', {'x': '2'}), False),
    ('z', ('1', {'x': '1', 'z': '0'}), True)
])
def test_greater_than_or_equal_column(compared: str, data, expected: bool):
    r = GreaterThanOrEqualColumn(compared)
    assert r._evaluator()(*r._prepare(data)) is expected

@pytest.mark.parametrize('compared,data,expected', [
    ('x', ('0', {'x': '1'}), True),
    ('x', ('1', {'x': '1'}), False),
    ('z', ('0', {'x': '0', 'z': '1'}), True)
])
def test_less_than_column(compared: str, data, expected: bool):
    r = LessThanColumn(compared)
    assert r._evaluator()(*r._prepare(data)) is expected


@pytest.mark.parametrize('compared,data,expected', [
    ('x', ('0', {'x': '1'}), True),
    ('x', ('1', {'x': '1'}), True),
    ('x', ('2', {'x': '1'}), False),
    ('z', ('0', {'x': '0', 'z': '1'}), True)
])
def test_less_than_or_equal_column(compared: str, data, expected: bool):
    r = LessThanOrEqualColumn(compared)
    assert r._evaluator()(*r._prepare(data)) is expected


@pytest.mark.parametrize('compare,row,expected', [
    ('col_b', ('', {'col_b': ''}), True),
    ('col_b', ('test', {'col_b': ''}), True),
    ('col_b', ('test', {'col_b': 'test'}), True),
    ('col_b', ('', {'col_b': 'test'}), False),
    (['col_b', 'col_c'], ('', {'col_b': '', 'col_c': ''}), True),
    (['col_b', 'col_c'], ('test', {'col_b': '', 'col_c': ''}), True),
    (['col_b', 'col_c'], ('test', {'col_b': '', 'col_c': 'test'}), True),
    (['col_b', 'col_c'], ('test', {'col_b': 'test', 'col_c': 'test'}), True),
    (['col_b', 'col_c'], ('', {'col_b': 'test', 'col_c': ''}), False),
    (['col_b', 'col_c'], ('', {'col_b': '', 'col_c': 'test'}), False),
    (['col_b', 'col_c'], ('', {'col_b': 'test', 'col_c': 'test'}), False)
])
def test_not_null_if_compare(compare, row, expected):
    r = NotNullIfCompare(compare)
    assert r._evaluator()(*r._prepare(row)) is expected


@pytest.mark.parametrize('other,row,expected', [
    ('col_b', ('', {'col_b': ''}), True),
    ('col_b', ('', {'col_b': 'x'}), True),
    ('col_b', ('x', {'col_b': ''}), True),
    ('col_b', ('x', {'col_b': 'x'}), False),
])
def test_other_cant_exist(other, row, expected):
    r = OtherCantExist(other)
    assert r._evaluator()(*r._prepare(row)) is expected


@pytest.mark.parametrize('other,row,expected', [
    ('col_b', ('', {'col_b': ''}), True),
    ('col_b', ('', {'col_b': 'x'}), True),
    ('col_b', ('x', {'col_b': ''}), False),
    ('col_b', ('x', {'col_b': 'x'}), True),
])
def test_other_must_exist(other, row, expected):
    r = OtherMustExist(other)
    assert r._evaluator()(*r._prepare(row)) is expected
